#!/usr/bin/env python3
#
# Copyright (C) 2021 Guillaume Bernard <contact@guillaume-bernard.fr>
#
# This is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
from typing import Dict, Any

from wikivents.models import ParticipatingEntity, ISO6391LanguageCode, EntityType, Event


class EventToDictEncoder:

    def __init__(self, event: Event, iso_639_1_language_code: ISO6391LanguageCode):
        self.event = event
        self.iso_639_1_language_code = iso_639_1_language_code

    def encode(self, participating_entities_ratio_to_keep: float = 0.5) -> Dict[str, Dict[str, Any]]:
        if participating_entities_ratio_to_keep < 0 or participating_entities_ratio_to_keep > 1.0:
            raise ValueError(f"The participating entities ratio to keep should be comprised between 0 and 1"
                             f" ({participating_entities_ratio_to_keep}) given.")
        return {
            self.event.id: {
                "iso_639_1_language_code": self.iso_639_1_language_code,
                "id": self.event.id,
                "type": EntityType.EVENT.name,
                "label": self.event.label(self.iso_639_1_language_code),
                "description": self.event.description(self.iso_639_1_language_code),
                "names": list(self.event.names(self.iso_639_1_language_code)),
                "processed_languages": list(self.event.processed_languages_iso_639_1_codes),
                "entities_kept_if_mentioned_in_more_at_least_X_languages":
                    self.event.nb_of_processed_languages * participating_entities_ratio_to_keep,
                "start": self.__get_start_date_or_none(),
                "end": self.__get_end_date_or_none(),
                "entities": {
                    "per": [
                        self.__get_participating_dict(participating_per, self.iso_639_1_language_code)
                        for participating_per in self.event.entities(
                            EntityType.PERSON, participating_entities_ratio_to_keep
                        )
                    ],
                    "gpe": [
                        self.__get_participating_dict(participating_gpe, self.iso_639_1_language_code)
                        for participating_gpe in self.event.entities(
                            EntityType.GPE, participating_entities_ratio_to_keep
                        )
                    ],
                    "org": [
                        self.__get_participating_dict(participating_org, self.iso_639_1_language_code)
                        for participating_org in self.event.entities(
                            EntityType.ORG, participating_entities_ratio_to_keep
                        )
                    ],
                }
            }
        }

    def __get_start_date_or_none(self) -> str:
        try:
            beginning = self.event.beginning.isoformat()
        except ValueError:
            beginning = str()
        return beginning

    def __get_end_date_or_none(self) -> str:
        try:
            end = self.event.end.isoformat()
        except ValueError:
            end = str()
        return end

    # pylint: disable=no-self-use
    # noinspection PyMethodMayBeStatic
    def __get_participating_dict(self, participating_entity: ParticipatingEntity,
                                 iso_639_1_language_code: ISO6391LanguageCode) -> Dict[str, Any]:
        return {
            "id": participating_entity.entity.id,
            "type": [entity_type.name for entity_type in participating_entity.entity.types()],
            "label": participating_entity.entity.label(iso_639_1_language_code),
            "description": participating_entity.entity.label(iso_639_1_language_code),
            "names": list(participating_entity.entity.names(iso_639_1_language_code)),
            "count": participating_entity.count
        }
